@@ -15,14 +15,15 @@
  */
 
 #include "globus_common.h"
+#include "globus_i_gram_protocol.h"
 #include "globus_gram_protocol_constants.h"
 
 /**
- * @defgroup globus_gram_protocol_error_messages Error Translation
+ * @defgroup globus_gram_protocol_error_messages Error Messages
  * @ingroup globus_gram_protocol_functions
  *
- * Functions in this section handle translating GRAM error codes to
- * strings which can help the user diagnose GRAM problems. 
+ * Functions in this section handle converting GRAM error codes to
+ * strings which can help the user diagnose GRAM problems.
  */
 static char *
 globus_l_gram_protocol_error_strings[GLOBUS_GRAM_PROTOCOL_ERROR_LAST] =
@@ -202,60 +203,223 @@ globus_l_gram_protocol_error_strings[GLOBUS_GRAM_PROTOCOL_ERROR_LAST] =
 /* 167 */     "the job is not running in the account named by the 'user_name' parameter."
 };
 
-static char *
-globus_l_gram_protocol_error_7_hack_message = GLOBUS_NULL;
+globus_thread_key_t                     globus_i_gram_protocol_error_key = 0;
+
+enum { GLOBUS_L_HACK_MESSAGE_MAX = 1024 };
 
 /**
- * Error code translation.
+ * @brief
+ * Get a description of a a GRAM error code
  * @ingroup globus_gram_protocol_error_messages
  *
- * This function takes the error code value and returns the associated error
- * code string. The string is statically allocated by the Globus GRAM Protocol
- * library and should not be modified or freed.
- *
+ * @details
+ * The @a globus_gram_protocol_error_string() function takes a GRAM error code
+ * value and returns the associated error code string for the message. The
+ * string is statically allocated by the GRAM Protocol
+ * library and should not be modified or freed by the caller. The string is
+ * intended to complete a sentence of the form
+ * "[operation] failed because ..."
+ * 
  * @param error_code
- *        The error code to look up.
- *
- * @return An error string containing the reason for the error. The error
- *         string is written to be used in the context
- *         "[operation] failed because [error_string]".
+ *        The error code to translate into a string.
  *
+ * @return
+ *     The @a globus_gram_protocol_error_string() function returns a static
+ *     string containing an explanation of the error.
  */
 const char *
-globus_gram_protocol_error_string(int error_code)
+globus_gram_protocol_error_string(
+    int                                 error_code)
 {
+    globus_hashtable_t *                hacktable;
+    char *                              str = NULL;
+
     if (error_code<0 || error_code>=GLOBUS_GRAM_PROTOCOL_ERROR_LAST)
         return("Invalid error code");
 
-    if ((error_code == GLOBUS_GRAM_PROTOCOL_ERROR_AUTHORIZATION)
-	&& globus_l_gram_protocol_error_7_hack_message)
-	return globus_l_gram_protocol_error_7_hack_message;
+    hacktable = globus_thread_getspecific(globus_i_gram_protocol_error_key);
 
-    return(globus_l_gram_protocol_error_strings[error_code]);
-} /* globus_gram_protocol_error_string() */
+    if (hacktable != NULL)
+    {
+        str = globus_hashtable_lookup(hacktable, (void *) error_code);
+    }
+
+    if (str == NULL)
+    {
+        str = globus_l_gram_protocol_error_strings[error_code];
+
+    }
+
+    return str;
+}
+/* globus_gram_protocol_error_string() */
 
 
 /**
- * GSI specific error code hack.
+ * @brief Replace the error message associated with error 7 with a custom message
  * @ingroup globus_gram_protocol_error_messages
  *
- * This function creates a custom version of the error message for the error
- * GLOBUS_GRAM_PROTOCOL_ERROR_AUTHORIZATION.  <b>This function should really
- * only used by the GRAM client library.</b>
+ * @details
+ * The @a globus_gram_protocol_error_7_hack_replace_message() function creates
+ * a custom version of the error message for the error
+ * @a GLOBUS_GRAM_PROTOCOL_ERROR_AUTHORIZATION. The string pointed to by the
+ * @a message parameter is copied to thread-local storage, and subsequent calls
+ * to @a globus_gram_protocol_error_string() with this error number will return
+ * this copy of the string. Each time
+ * globus_gram_protocol_error_7_hack_replace_message() is called for
+ * a particular thread, the previous message is freed.
+ *
+ * The purpose of this function is to allow more meaningful error messages to
+ * be generated when authentication failures occur. In particular, the specific
+ * GSSAPI error reason can be used in place of a generic authorization failure
+ * message.
  *
  * @param message
- *        The new message to be associated with error code 7.
+ *        The new message to be associated with the
+ *        GLOBUS_GRAM_PROTOCOL_ERROR_AUTHORIZATION error code.
+ *
+ * @note
+ *     Since Globus 5.0.0, this function uses thread-specific storage, so that
+ *     the value returned by globus_gram_protocol_error_string() for 
+ *     GLOBUS_GRAM_PROTOCOL_ERROR_AUTHORIZATION is that for the last
+ *     authorization error where
+ *     @a globus_gram_protocol_error_7_hack_replace_message() was called from
+ *     this thread.
  */
 void
 globus_gram_protocol_error_7_hack_replace_message(const char * message)
 {
-    if (globus_l_gram_protocol_error_7_hack_message)
-	globus_libc_free(globus_l_gram_protocol_error_7_hack_message);
+    globus_i_gram_protocol_error_hack_replace_message(7, message);
+}
+/* globus_gram_protocol_error_7_hack_replace_message() */
 
-    globus_l_gram_protocol_error_7_hack_message = GLOBUS_NULL;
-
-    if (message)
-	globus_l_gram_protocol_error_7_hack_message = 
-	    globus_libc_strdup(message);
+/**
+ * @brief Replace the error message associated with error 10 with a custom message
+ * @ingroup globus_gram_protocol_error_messages
+ *
+ * @details
+ * The @a globus_gram_protocol_error_10_hack_replace_message() function creates
+ * a custom version of the error message for the error
+ * @a GLOBUS_GRAM_PROTOCOL_ERROR_PROTOCOL_FAILED. The string pointed to by the
+ * @a message parameter is copied to thread-local storage, and subsequent calls
+ * to @a globus_gram_protocol_error_string() with this error number will return
+ * this copy of the string. Each time
+ * globus_gram_protocol_error_10_hack_replace_message() is called for
+ * a particular thread, the previous message is freed.
+ *
+ * The purpose of this function is to allow more meaningful error messages to
+ * be generated when protocol errors occur. In particular, the specific
+ * XIO error reason can be used in place of a generic protocol failure
+ * message.
+ *
+ * @param message
+ *        The new message to be associated with the
+ *        GLOBUS_GRAM_PROTOCOL_ERROR_PROTOCOL_FAILED error code.
+ *
+ * @note
+ *     Since Globus 5.0.0, this function uses thread-specific storage, so that
+ *     the value returned by globus_gram_protocol_error_string() for 
+ *     GLOBUS_GRAM_PROTOCOL_ERROR_PROTOCOL_FAILED is that for the last
+ *     authorization error where
+ *     @a globus_gram_protocol_error_10_hack_replace_message() was called from
+ *     this thread.
+ */
+void
+globus_gram_protocol_error_10_hack_replace_message(const char * message)
+{
+    globus_i_gram_protocol_error_hack_replace_message(10, message);
 }
 
+#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
+/**
+ * Replace an error code context-specific error message with a new message
+ *
+ * @param errorcode
+ *     Integer error code to associate the context-specific message with
+ * @param message
+ *     The new message string.
+ */
+void
+globus_i_gram_protocol_error_hack_replace_message(
+    int                                 errorcode,
+    const char *                        message)
+{
+    globus_hashtable_t *                hacktable;
+    char *                              old;
+    int                                 rc;
+
+    hacktable = globus_thread_getspecific(globus_i_gram_protocol_error_key);
+    if (!hacktable)
+    {
+        hacktable = malloc(sizeof(globus_hashtable_t));
+
+        if (hacktable != NULL)
+        {
+            rc = globus_hashtable_init(
+                    hacktable,
+                    17,
+                    globus_hashtable_int_hash,
+                    globus_hashtable_int_keyeq);
+            if (rc != GLOBUS_SUCCESS)
+            {
+                free(hacktable);
+                hacktable = NULL;
+            }
+
+            globus_thread_setspecific(
+                globus_i_gram_protocol_error_key, hacktable);
+        }
+    }
+
+    if (hacktable == NULL)
+    {
+        return;
+    }
+
+    old = globus_hashtable_remove(hacktable, (void *) errorcode);
+    if (old)
+    {
+        free(old);
+        old = NULL;
+    }
+
+    if (message)
+    {
+        old = strdup(message);
+    }
+
+    if (old)
+    {
+        rc = globus_hashtable_insert(
+                hacktable,
+                (void *) errorcode,
+                old);
+
+        if (rc != GLOBUS_SUCCESS)
+        {
+            free(old);
+        }
+    }
+}
+/* globus_l_gram_error_hack_replace_message() */
+
+static
+void
+globus_l_gram_protocol_errors_destroy(
+    void *                              str)
+{
+    if (str)
+    {
+        free(str);
+    }
+}
+/* globus_l_gram_protocol_errors_destroy() */
+
+void
+globus_i_gram_protocol_error_destroy(
+    void *                              arg)
+{
+    globus_hashtable_destroy_all(arg, globus_l_gram_protocol_errors_destroy);
+    free(arg);
+}
+#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */
